''{********PROP SD LOADER*********}
' v1.0 - 13 June 2006 - original version
' v1.1 - 22 June 2008 - update to support SD Loading

''_____________________________________________________________________________
''
''This object lets a Propeller chip load up another Propeller chip in the same
''way the PC normally does.                                                    
''We assume the following:
''pins 0..2 are connected to RES,p31,p30
''pins 8..11 are connected to an SD card
''Version of chip is the Prop I
''we want to loadram only, not eeprom
''
''OBJ loader : "PropSDLoader"
''
''
''PUB LoadPropeller
''
''  loader.Connect(0, 1, 2, 1, loader#LoadRun, 8, string("file.bin"))
''
''
'' FINAL OBJECT DRIVES RES NOT RESn VIA NPN 

CON

  #0, LoadSuccess,ErrorConnect, ErrorVersion, ErrorChecksum, ErrorProgram, ErrorVerify, ErrorSD
  #0, Shutdown, LoadRun, ProgramShutdown, ProgramRun
VAR

  long P31, P30, LFSR, Ver, Echo, ByteCount, RemBytes, BytesLeft 
dat
CurrentRead byte 0[16] 

OBJ
  sdfat: "fsrw_with_Kill"
PUB Connect(PinRESn, PinP31, PinP30, Version, Command, SDPin,FileString) : Error
{*********************** Lets get the pins and variables ready ************************ }
  bytesLeft := 16                                                               'init variables
  RemBytes := 0
  ByteCount := 0
  P31 := PinP31                                                                 'set our p31 variable
  P30 := PinP30                                                                 'set our p30 variable
  dira[PinRESn] := 1                                                            'set reset pin to output
  dira[PinP31] := 1                                                             'set p31 to output
  dira[PinP30] := 0                                                             'set p30 to input
  outa[PinRESn] := 0                                                            'set reset pin low   
  outa[PinP31] := 1                                                             'set p31 (TX) high

{****************************** Lets get the data ready ******************************* }
  
  if Error := \sdfat.mount(SDPin)                                               'mount card
    sdfat.kill
    return Error                                                                'we failed to mount card, could be a critical error, dont retry
                
  waitcnt(clkfreq / 10 + cnt)                                                   'wait 100ms, helps eliminate lexar SD card errors
  if Error := \sdfat.popen(FileString, "r")                                     'open the file
    sdfat.kill
    return Error                                                                'we failed to open file, could be a critical error, check error code to check if we should retry
 
  sdfat.pread(@CurrentRead, 16)                                                 'get the first 16 bytes
  ByteCount := byte[@CurrentRead][8] | byte[@CurrentRead][9] << 8               'figure out how many bytes the file is
  RemBytes := ByteCount - 16                                                    'how many bytes do we still need to read?

{****************************** Lets fire this baby up! ******************************* } 

  outa[PinRESn] := 1                                                            'bring the prop online
  waitcnt(clkfreq / 10 + cnt)                                                   'stabalization time
  
{****************************** Prop should be ready for data************************** }

  if Error := \Communicate(Version, Command)
    sdfat.kill
    return Error                                                                'we failed to load it, return error and try again?
  else
   sdfat.kill
   return loadsuccess                                                           'we loaded it correctly, return success
PRI Communicate(Version, Command)

  'output calibration pulses
  BitsOut(%01, 2)               

  'send LFSR pattern
  LFSR := "P"                   
  repeat 250
    BitsOut(IterateLFSR, 1)

  'receive and verify LFSR pattern
  repeat 250                   
    if WaitBit(1) <> IterateLFSR
      abort ErrorConnect

  'receive chip version      
  repeat 8
    Ver := WaitBit(1) << 7 + Ver >> 1

  'if version mismatch, shutdown and abort
  if Ver <> Version
    BitsOut(Shutdown, 32)
    abort ErrorVersion

  'send command
  BitsOut(Command, 32)

  'handle command details
  if Command          

    'send long count
    'ByteCount := byte[CodePtr][8] | byte[CodePtr][9] << 8   
    BitsOut(ByteCount >> 2, 32)

    'send bytes
    repeat ByteCount
      if bytesleft
       BitsOut(byte[@CurrentRead+(16-bytesleft)], 8)
       bytesleft--
      else
       sdfat.pread(@CurrentRead, 16)
       BitsOut(byte[@CurrentRead], 8)
       bytesleft := 15 
    'allow 250ms for positive checksum response
    if WaitBit(25)
      abort ErrorChecksum

    'eeprom program command
    if Command > 1
    
      'allow 5s for positive program response
      if WaitBit(500)
        abort ErrorProgram
        
      'allow 2s for positive verify response
      if WaitBit(200)
        abort ErrorVerify
                

PRI IterateLFSR : Bit

  'get return bit
  Bit := LFSR & 1
  
  'iterate LFSR (8-bit, $B2 taps)
  LFSR := LFSR << 1 | (LFSR >> 7 ^ LFSR >> 5 ^ LFSR >> 4 ^ LFSR >> 1) & 1
  

PRI WaitBit(Hundredths) : Bit | PriorEcho

  repeat Hundredths
  
    'output 1t pulse                        
    BitsOut(1, 1)
    
    'sample bit and echo
    Bit := ina[P30]
    PriorEcho := Echo
    
    'output 2t pulse
    BitsOut(0, 1)
    
    'if echo was low, got bit                                      
    if not PriorEcho
      return
      
    'wait 10ms
    waitcnt(clkfreq / 100 + cnt)

  'timeout, abort
  abort ErrorConnect

  
PRI BitsOut(Value, Bits)

  repeat Bits

    if Value & 1
    
      'output '1' (1t pulse)
      outa[P31] := 0                        
      Echo := ina[P30]
      outa[P31] := 1
      
    else
    
      'output '0' (2t pulse)
      outa[P31] := 0
      outa[P31] := 0
      Echo := ina[P30]
      Echo := ina[P30]
      outa[P31] := 1

    Value >>= 1
    